package com.nts.isip.tcp;

import com.nts.isip.tcp.exceptions.CheckSumException;
import com.nts.isip.tcp.exceptions.FrameSourceException;
import com.nts.isip.tcp.isipgui.ServerConfigDialog;
import com.nts.isip.tcp.listeners.StreamStateListener;
import com.nts.isip.tcp.processors.ParsingFrameProcessor;
import com.nts.isip.tcp.sources.FrameSource;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.IOException;

import java.util.Properties;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;


/**
 * Client control class. Handles commands from user. It establishes required connections
 * to server, creates a container with various Action objects for available user actions
 * (e.g. start/stop calculation stream, connect/disconnect from server, etc.) and passes
 * it to view to associate actions with user interface buttons. It initializes and keeps
 * reference to the stream thread (StreamThread), which is responsible for pumping frame
 * images and data from server when stream is started. Commands are executed by various
 * helper inner classes, extending the Action class.
 */
public class ClientCtrl {
    private StreamThread _streamThr;
    private FrameSource _strSource;

    //private ClientView _view;
    private ClientFrame _parentFrame;
    private ConnectAction _connect = new ConnectAction();
    private DisconnectAction _disconnect = new DisconnectAction();
    private ReprocessAction _reprocess = new ReprocessAction();

    public ClientCtrl() {
        //_view = new IcamPropertiesPanel();
        _parentFrame = new ClientFrame();
        _streamThr = new StreamThread(_strSource, _parentFrame,
                new ParsingFrameProcessor());
        //_view = _streamThr.getView();
        //_view.getTopFrame().setDefaultCloseOperation(
        //		javax.swing.JFrame.EXIT_ON_CLOSE);
        _streamThr.start();

        _connect.setEnabled(true);
        _disconnect.setEnabled(false);

        java.util.Map<Integer, Action> cmdMap = new java.util.HashMap<Integer, Action>();
        cmdMap.put(CmdId.START_CALC_STREAM, new StartStreamAction(false));
        cmdMap.put(CmdId.STOP_STREAM, new StopStreamAction());
        cmdMap.put(CmdId.CONNECT, _connect);
        cmdMap.put(CmdId.DISCONNECT, _disconnect);
        cmdMap.put(CmdId.FIRST_FRAME, new StreamFrameAction("<<", "first"));
        cmdMap.put(CmdId.PREV_FRAME, new StreamFrameAction("<", "prev"));
        cmdMap.put(CmdId.NEXT_FRAME, new StreamFrameAction(">", "next"));
        cmdMap.put(CmdId.LAST_FRAME, new StreamFrameAction(">>", "last"));
        cmdMap.put(CmdId.RELOAD_FRAME, new StreamFrameAction(".", "reload"));
        cmdMap.put(CmdId.BY_TIME_FRAME,
            new StreamFrameAction("by time", "byTime"));
        cmdMap.put(CmdId.STARTSTOP_FRAME_RECORDING,
            new StartStopFrameRecording("Start rec"));
        cmdMap.put(CmdId.REPROCESS, _reprocess);
        cmdMap.put(CmdId.SYSTEM_STORE, new SystemStoreAction());
        cmdMap.put(CmdId.TEMPLATE_STORE, new TemplateStoreAction());

        _parentFrame.init(cmdMap, _streamThr);
        connect();
    }

    private void connect() {
        ServerConfigDialog confDlg = new ServerConfigDialog(_parentFrame.getTopFrame());
        confDlg.setLocationRelativeTo(_parentFrame.getTopFrame());
        confDlg.setVisible(true);

        if (_strSource != null) {
            stopStream();

            try {
                _strSource.disconnect();
            } catch (java.io.IOException ex) {
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "Can not disconnect. The system told us:\n" +
                    ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }
        }

        if (!confDlg.getConnected()) {
            _parentFrame.setStatus("Not connected");
            _connect.setEnabled(true);
            _disconnect.setEnabled(false);
        } else {
            try {
                _parentFrame.setStatus("Connecting...");
                _strSource = confDlg.getFrameSource();
                //_strSource.isDirEmpty();
                _streamThr.setStreamSource(_strSource);

                //if(_strSource instanceof IcamFrameSource) {
                String title = _parentFrame.getTopFrame().getTitle();
                _parentFrame.setICamPanels();
                //} else {
                //	((ClientFrame)_view).setMainPanel();
                //	_streamThr.setView(((ClientFrame)_view).getView());
                //}

                //_view = _streamThr.getView();
                _parentFrame.getTopFrame()
                            .setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
                _parentFrame.getTopFrame()
                            .setTitle(title + " " +
                    confDlg.getFrameSourceTitle());

                _parentFrame.setStatus("Connected");
                _connect.setEnabled(false);
                _disconnect.setEnabled(true);
            } catch (IOException e) {
                String msg = "Could not connect to server.\n" +
                    "The system told us: " + e.getMessage();
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(), msg,
                    "Error", JOptionPane.ERROR_MESSAGE);

                _parentFrame.setStatus("Not connected");
                _connect.setEnabled(true);
                _disconnect.setEnabled(false);
            } catch (NullPointerException e) {
                //e.printStackTrace();
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "Enter data for connecting or check if the entered data are correct!",
                    "Information", JOptionPane.INFORMATION_MESSAGE);
                _parentFrame.setStatus("Not connected");
                _connect.setEnabled(true);
                _disconnect.setEnabled(false);
            } /*catch (FrameSourceException e) {
              JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                  e.getMessage(), "Information",
                  JOptionPane.INFORMATION_MESSAGE);
              _parentFrame.setStatus("Not connected");
              _connect.setEnabled(true);
              _disconnect.setEnabled(false);
            }*/}
    }

    private void startStream(boolean calc) {
        _streamThr.startStream(calc);
    }

    private void stopStream() {
        _streamThr.stopStream();
    }

    private class StartStreamAction extends FrameStreamControlAction
        implements StreamStateListener {
        private boolean _calc;

        public StartStreamAction(boolean calc) {
            super("Start");
            this.setEnabled(false);
            _calc = calc;
            _streamThr.addStateListener(this);
            _streamThr.addStreamSourceStateListener(this);
        }

        public void actionPerformed(ActionEvent e) {
            startStream(_calc);
        }

        public void streamStateChanged(boolean active) {
            this.setEnabled(!active);
        }
    }

    private class StopStreamAction extends FrameStreamControlAction
        implements StreamStateListener {
        public StopStreamAction() {
            super("Stop");
            _streamThr.addStateListener(this);
            this.setEnabled(false);
        }

        public void actionPerformed(ActionEvent e) {
            stopStream();
            this.setEnabled(false);
        }

        public void streamStateChanged(boolean active) {
            this.setEnabled(active);
        }
    }

    private class ConnectAction extends AbstractAction {
        public ConnectAction() {
            super("Connect");
        }

        public void actionPerformed(ActionEvent e) {
            stopStream();
            connect();
        }
    }

    private class DisconnectAction extends AbstractAction {
        public DisconnectAction() {
            super("Disconnect");
        }

        public void actionPerformed(ActionEvent e) {
            try {
                stopStream();
                _strSource.disconnect();
            } catch (java.io.IOException ex) {
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "Can not disconnect. The system told us:\n" +
                    ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }

            _connect.setEnabled(true);
            _disconnect.setEnabled(false);
            _parentFrame.setStatus("Disconnected");

            Properties prop = new Properties();

            try {
                java.io.InputStream in = getClass().getClassLoader()
                                             .getResource("version.properties")
                                             .openStream();
                prop.load(in);
            } catch (IOException ioe) {
            }

            if (prop.getProperty("version") != null) {
                _parentFrame.getTopFrame()
                            .setTitle("isip TCP client version " +
                    prop.getProperty("version"));
            } else {
                _parentFrame.getTopFrame().setTitle("isip TCP client");
            }
        }
    }

    private class ReprocessAction extends AbstractAction {
        private String _label;

        public ReprocessAction() {
            super("Reprocess");
            _label = "Reprocess";
        }

        public void actionPerformed(ActionEvent e) {
            if (_label.endsWith("Reprocess")) {
                ServerConfigDialog confDlg = new ServerConfigDialog(_parentFrame.getTopFrame());
                confDlg.setLocationRelativeTo(_parentFrame.getTopFrame());
                confDlg.setVisible(true);

                if (confDlg.getConnected()) {
                    try {
                        _streamThr.setAuxillarySource(confDlg.getFrameSource());
                        //_streamThr.getAuxillarySource().isDirEmpty();
                        _label = "Reprocess off";
                        this.putValue(Action.NAME, _label);
                    } catch (IOException ex) {
                        String msg = "Could not connect to server.\n" +
                            "The system told us: " + ex.getMessage();
                        JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                            msg, "Error", JOptionPane.ERROR_MESSAGE);
                        _streamThr.setAuxillarySource(null);
                    } catch (NullPointerException ex) {
                        JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                            "Enter data for connecting!", "Information",
                            JOptionPane.INFORMATION_MESSAGE);
                        _streamThr.setAuxillarySource(null);
                    } /*catch (FrameSourceException ex) {
                      JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                          ex.getMessage(), "Information",
                          JOptionPane.INFORMATION_MESSAGE);
                      _streamThr.setAuxillarySource(null);
                    }*/}
            } else {
                _streamThr.setAuxillarySource(null);
                _label = "Reprocess";
                this.putValue(Action.NAME, _label);
            }
        }
    }

    private class StreamFrameAction extends FrameStreamControlAction
        implements StreamStateListener {
        String _name;
        String _cmd;

        StreamFrameAction(String name, String cmd) {
            super(name);
            this.setEnabled(false);
            _name = name;
            _cmd = cmd;
            _streamThr.addStreamSourceStateListener(this);
            _streamThr.addStateListener(this);
        }

        public void actionPerformed(ActionEvent e) {
            try {
                if (_cmd.equals("first")) {
                    _streamThr.getFirstFrame();
                } else if (_cmd.equals("prev")) {
                    _streamThr.getPrevFrame();
                } else if (_cmd.equals("next")) {
                    _streamThr.getNextFrame();
                } else if (_cmd.equals("last")) {
                    _streamThr.getLastFrame();
                } else if (_cmd.equals("reload")) {
                    _streamThr.reloadFrame();
                } else if (_cmd.equals("byTime")) {
                    final JDialog dlg = new JDialog(_parentFrame.getTopFrame());
                    dlg.setLayout(new BorderLayout());
                    dlg.setSize(new Dimension(200, 100));
                    dlg.setLocationRelativeTo(_parentFrame.getTopFrame());

                    JPanel timePan = new JPanel();
                    timePan.setLayout(new GridBagLayout());

                    GridBagConstraints c = new GridBagConstraints();
                    c.gridx = 0;
                    c.gridy = 0;
                    c.fill = GridBagConstraints.HORIZONTAL;
                    c.insets = new java.awt.Insets(5, 5, 5, 5);
                    timePan.add(new JLabel("Timestamp: ", JLabel.LEFT), c);

                    c.gridx = 0;
                    c.gridy = 1;
                    c.weightx = .6;

                    final JTextField keyField = new JTextField(12);
                    timePan.add(keyField, c);

                    JButton button = new JButton("OK");
                    button.addActionListener(new ActionListener() {
                            public void actionPerformed(ActionEvent e) {
                                if (!keyField.getText().equals("")) {
                                    try {
                                        int timestamp = Integer.valueOf(keyField.getText())
                                                               .intValue();
                                        _streamThr.getFrameByTime(timestamp);
                                    } catch (java.io.IOException ex) {
                                        if (ex.getMessage() != null) {
                                            JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                                "Error transmitting package.\n" +
                                                "The system told us: " +
                                                ex.getMessage(), "Error",
                                                JOptionPane.ERROR_MESSAGE);
                                        } else {
                                            JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                                "Error transmitting package.\n",
                                                "Error",
                                                JOptionPane.ERROR_MESSAGE);
                                        }
                                    } /*catch (ServerError ex) {
                                      JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                                      "Error transmitting package.\n" + "The system told us: " +
                                              ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
                                    } catch (ProtocolError ex) {
                                      JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                              "Error transmitting package.\n" + "The system told us: " +
                                              ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
                                    }*/
                                    catch (CheckSumException ex) {
                                        JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                            "A CheckSum Error has occured",
                                            "Error", JOptionPane.ERROR_MESSAGE);
                                    } catch (NumberFormatException ex) {
                                        JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                            "A Number format error has occured",
                                            "Error", JOptionPane.ERROR_MESSAGE);
                                    }
                                }

                                dlg.dispose();
                            }
                        });

                    JPanel btnPan = new JPanel();
                    btnPan.setLayout(new BoxLayout(btnPan, BoxLayout.X_AXIS));
                    btnPan.setBorder(BorderFactory.createEmptyBorder(0, 10, 10,
                            10));
                    btnPan.add(Box.createHorizontalGlue());
                    btnPan.add(button);

                    dlg.getContentPane()
                       .setLayout(new BoxLayout(dlg.getContentPane(),
                            BoxLayout.Y_AXIS));

                    dlg.getContentPane().add(timePan);
                    dlg.getContentPane().add(Box.createVerticalGlue());
                    dlg.add(btnPan);
                    dlg.pack();

                    dlg.setVisible(true);
                }
            } catch (java.io.IOException ex) {
                if (ex.getMessage() != null) {
                    JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                        "Error transmitting package.\n" +
                        "The system told us: \n" + ex.getMessage(), "Error",
                        JOptionPane.ERROR_MESSAGE);
                } else {
                    JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                        "Error transmitting package.\n", "Error",
                        JOptionPane.ERROR_MESSAGE);
                }
            } /*catch (ProtocolError ex) {
              JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                  "Error transmitting package.\n" + "The system told us: " +
                  ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            } catch (ServerError ex) {
              JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                  "Error transmitting package.\n" + "The system told us: " +
                  ex.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            }*/
            catch (IllegalArgumentException ex) {
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "A data error has occured.\n The Processor requires" +
                    " images of type 8-bit, 16-but grayscale or Color images",
                    "Error", JOptionPane.ERROR_MESSAGE);
            } catch (CheckSumException ex) {
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "A CheckSum Error has occured", "Error",
                    JOptionPane.ERROR_MESSAGE);
            }
        }

        public void streamStateChanged(boolean active) {
            this.setEnabled(!active);
        }
    }

    private class StartStopFrameRecording extends AbstractAction {
        private String _label;

        public StartStopFrameRecording(String label) {
            super(label);
            _label = label;
        }

        public void actionPerformed(ActionEvent e) {
            if (_label.endsWith("Start rec")) {
                ServerConfigDialog confDlg = new ServerConfigDialog(null);
                confDlg.setLocationRelativeTo(null);
                confDlg.setVisible(true);

                FrameSource frameSource = null;

                if (confDlg.getConnected()) {
                    try {
                        frameSource = confDlg.getFrameSource();
                        _streamThr.setTargetFrameSource(frameSource);
                        _streamThr.setFrameRecording(true);
                        _label = "Stop rec";
                        this.putValue(Action.NAME, _label);
                    } catch (IOException ex) {
                        String msg = "Could not connect to server.\n" +
                            "The system told us: " + ex.getMessage();
                        JOptionPane.showMessageDialog(null, msg, "Error",
                            JOptionPane.ERROR_MESSAGE);
                        frameSource = null;
                    } catch (NullPointerException ex) {
                        JOptionPane.showMessageDialog(null,
                            "Enter data for connecting!", "Information",
                            JOptionPane.INFORMATION_MESSAGE);
                        frameSource = null;
                    }
                }
            } else if (_label.equals("Stop rec")) {
                _streamThr.getTargetFrameSource().stopWriting();
                _streamThr.setTargetFrameSource(null);
                _streamThr.setFrameRecording(false);
                _label = "Start rec";
                this.putValue(Action.NAME, _label);
            }
        }
    }

    private class SystemStoreAction extends AbstractAction {
        public SystemStoreAction() {
            super("SystemStore");
        }

        public void actionPerformed(ActionEvent e) {
            try {
                _streamThr.systemStore();
            } catch (IOException ex) {
                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                    "Could not store system settings.\n The system told us: " +
                    ex.getMessage(), "Information",
                    JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    private class TemplateStoreAction extends AbstractAction {
        public TemplateStoreAction() {
            super("TemplateStore");
        }

        public void actionPerformed(ActionEvent e) {
            final JDialog dlg = new JDialog(_parentFrame.getTopFrame());
            dlg.setLayout(new BorderLayout());
            dlg.setSize(new Dimension(200, 100));
            dlg.setLocationRelativeTo(_parentFrame.getTopFrame());

            JPanel storePan = new JPanel();
            storePan.setLayout(new GridBagLayout());

            GridBagConstraints c = new GridBagConstraints();
            c.gridx = 0;
            c.gridy = 0;
            c.fill = GridBagConstraints.HORIZONTAL;
            c.insets = new java.awt.Insets(5, 5, 5, 5);
            //storePan.add(new JLabel("ID of the template: ", JLabel.LEFT), c);
            storePan.add(new JLabel("Please enter a number: ", JLabel.LEFT), c);

            c.gridx = 0;
            c.gridy = 1;
            c.weightx = .6;

            final JTextField keyField = new JTextField(12);
            storePan.add(keyField, c);

            JButton button = new JButton("OK");
            button.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        try {
                            if (!keyField.getText().equals("")) {
                                String key = keyField.getText();

                                try {
                                    _streamThr.templateStore(key);
                                } catch (FrameSourceException ex) {
                                    int selected = JOptionPane.showConfirmDialog(_parentFrame.getTopFrame(),
                                            "Error while trying to save the tamplate. " +
                                            "The system told us:\n" +
                                            ex.getMessage() +
                                            "\n\nDo you want to delete it?");

                                    if (selected == JOptionPane.YES_OPTION) {
                                        String res = _streamThr.deleteTemplate(key);

                                        if (res.equals("S_OK")) {
                                            try {
                                                _streamThr.templateStore(key);
                                            } catch (FrameSourceException fex) {
                                                JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                                    "Could not store the template.\n The system told us: " +
                                                    fex.getMessage(), "Error",
                                                    JOptionPane.ERROR_MESSAGE);
                                            }
                                        } else {
                                            JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                                "Could not delete the template.\n The system told us: " +
                                                res, "Error",
                                                JOptionPane.ERROR_MESSAGE);
                                        }
                                    }
                                }
                            }

                            dlg.dispose();
                        } catch (IOException ex) {
                            JOptionPane.showMessageDialog(_parentFrame.getTopFrame(),
                                "Could not store system settings.\n The system told us: " +
                                ex.getMessage(), "Information",
                                JOptionPane.INFORMATION_MESSAGE);
                        }
                    }
                });

            JPanel btnPan = new JPanel();
            btnPan.setLayout(new BoxLayout(btnPan, BoxLayout.X_AXIS));
            btnPan.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
            btnPan.add(Box.createHorizontalGlue());
            btnPan.add(button);

            dlg.getContentPane()
               .setLayout(new BoxLayout(dlg.getContentPane(), BoxLayout.Y_AXIS));

            dlg.getContentPane().add(storePan);
            dlg.getContentPane().add(Box.createVerticalGlue());
            dlg.add(btnPan);
            dlg.pack();

            dlg.setVisible(true);
        }
    }
}
