package com.nts.isip.tcp;

import com.nts.isip.tcp.exceptions.CheckSumException;
import com.nts.isip.tcp.streams.IntelDataInputStream;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class DataPackageParser {
    private final static int MAGICCHECKNUM = 0x12345678;
    private final static int TWODDATAARRAYS = 30;
    private Vector<ImageRepresentation> _images;
    private Vector<HistogramRepresentation> _histograms;
    private double[][] _points;
    private float[] _profile;
    private float[] _fit;
    private int _fitPts;
    private int _msrX;
    private int _msrY;
    private boolean _hasMeasurePoints;
    private int _timeStamp;
    private int _frameCnt;
    private double[] _numericData;
    private int _errCode;
    private Hashtable<String, SectionProperties> _secProps;
    private byte[] _fullPackage;

    public void parse(byte[] data) throws IOException, CheckSumException {
        _images = new Vector<ImageRepresentation>();
        _histograms = new Vector<HistogramRepresentation>();
        _points = new double[TWODDATAARRAYS][];
        _profile = null;
        _fit = null;
        _fitPts = 0;
        _msrX = 0;
        _msrY = 0;
        _hasMeasurePoints = false;
        _timeStamp = 0;
        _frameCnt = 0;
        _numericData = null;
        _secProps = new Hashtable<String, SectionProperties>();
        _fullPackage = data;

        Properties prop = new Properties();

        try {
            java.io.FileInputStream in = new java.io.FileInputStream(
                    "isip_client_chart.properties");

            try {
                prop.load(in);
            } finally {
                in.close();
            }
        } catch (IOException e) {
        }

        Enumeration en = prop.keys();

        while (en.hasMoreElements()) {
            String key = (String) en.nextElement();

            if (key.startsWith("ser") && key.endsWith(".PKGSECID")) {
                int number = 0;
                Pattern p = Pattern.compile(".*?(\\d+).*?");
                Matcher m = p.matcher(key);

                if (m.find()) {
                    number = Integer.valueOf(m.group(1)).intValue();
                }

                String secId = prop.getProperty("ser" + number + ".PKGSECID");
                SectionProperties secProps = new SectionProperties();
                secProps.pkgSecType = prop.getProperty("ser" + number +
                        ".PKGSECTYPE");
                secProps.pkgTabIdx = Integer.valueOf(prop.getProperty("ser" +
                            number + ".tabidx")).intValue();

                try {
                    secProps.idx = Integer.valueOf(prop.getProperty("ser" +
                                number + ".idx").trim()).intValue();
                } catch (NumberFormatException e) {
                    secProps.idx = -1;
                }

                secProps.color = prop.getProperty("ser" + number + ".color");

                try {
                    secProps.lines = Integer.valueOf(prop.getProperty("ser" +
                                number + ".lines").trim()).intValue();
                } catch (NumberFormatException e) {
                    secProps.lines = 0;
                }

                secProps.name = prop.getProperty("ser" + number + ".name");
                secProps.type = prop.getProperty("ser" + number + ".type");
                _secProps.put(secId, secProps);
            }
        }

        ByteArrayInputStream in = new ByteArrayInputStream(data);
        IntelDataInputStream inNum = new IntelDataInputStream(in);

        int magic = inNum.readIntelInt();
        int size = inNum.readIntelInt();
        int infoSize = inNum.readIntelInt();
        int sectnCnt = inNum.readIntelInt();
        int lastSecOff = inNum.readIntelInt();
        int checkSum = inNum.readIntelInt();

        int check = 0;
        check = check ^ magic ^ size ^ infoSize ^ sectnCnt ^ lastSecOff ^
            checkSum;

        if (check != MAGICCHECKNUM) {
            throw new CheckSumException("Check sum error");
        }

        int leftbytes = size - infoSize;
        int offset = 0 + infoSize;
        int checkSection;

        while (sectnCnt > 0) {
            in = new ByteArrayInputStream(data, offset, leftbytes);
            inNum = new IntelDataInputStream(in);

            String tag = inNum.readIntelString(4);
            int sectionSize = inNum.readIntelInt();

            if (sectionSize > leftbytes) {
                throw new CheckSumException("Section size error");
            }

            int sectionHdrSize = inNum.readIntelInt();
            int nextSectionOffset = inNum.readIntelInt();
            int sectionCheckSum = inNum.readIntelInt();

            byte[] tagBytes = tag.getBytes();
            int tagInt = tagBytes[0] + (tagBytes[1] * (1 << 8)) +
                (tagBytes[2] * (1 << 16)) + (tagBytes[3] * (1 << 24));

            checkSection = 0;
            checkSection = checkSection ^ tagInt ^ sectionSize ^
                sectionHdrSize ^ nextSectionOffset ^ sectionCheckSum;

            if (checkSection != MAGICCHECKNUM) {
                throw new CheckSumException("Check sum error");
            }

            if (_secProps.containsKey(tag) &&
                    _secProps.get(tag).pkgSecType.equals("POINTS")) {
                _points[_secProps.get(tag).idx] = readPointsTypeSec(inNum);
                leftbytes -= sectionSize;
                offset += nextSectionOffset;
            } else if (_secProps.containsKey(tag) &&
                    _secProps.get(tag).pkgSecType.equals("IMAGE")) {
                ImageRepresentation image = readImgSec(data, inNum,
                        offset + sectionHdrSize, sectionSize - sectionHdrSize);
                image.tabIndex = _secProps.get(tag).pkgTabIdx;
                image.idx = _secProps.get(tag).idx;

                _images.add(image);
                leftbytes -= sectionSize;
                offset += nextSectionOffset;
            } else if (_secProps.containsKey(tag) &&
                    _secProps.get(tag).pkgSecType.equals("HISTOGRAM")) {
                if (tag.equals("HSRL")) {
                    HistogramRepresentation histogram = readHistSec(inNum);
                    histogram.tabIndex = _secProps.get(tag).pkgTabIdx;
                    histogram.idx = _secProps.get(tag).idx;
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;

                    if (!_secProps.contains("PLGT")) {
                        _histograms.add(histogram);
                    }
                } else {
                    HistogramRepresentation histogram;

                    if (tag.equals("PLGT")) {
                        histogram = readHistPointSec(inNum);
                    } else {
                        histogram = readHistSec(inNum);
                    }

                    histogram.tabIndex = _secProps.get(tag).pkgTabIdx;
                    histogram.idx = _secProps.get(tag).idx;
                    _histograms.add(histogram);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                }
            } else {
                if (tag.equals("FRMI")) {
                    readFRMIsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("TRAC")) {
                    if (sectionSize > 40) {
                        readNewTRACsec(inNum);
                    } else {
                        readTRACsec(inNum);
                    }

                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("DCNT")) {
                    readDCNTsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("PROF")) {
                    readPROFsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("IPRF")) {
                    readIPRFsec(data, inNum, offset + sectionHdrSize,
                        sectionSize - sectionHdrSize);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("FITP")) {
                    readFITPsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("PLNS")) {
                    readPLNSsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else if (tag.equals("STAT")) {
                    readSTTSsec(inNum);
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                } else {
                    leftbytes -= sectionSize;
                    offset += nextSectionOffset;
                }
            }

            sectnCnt = sectnCnt - 1;
        }

        if (_errCode != 0) {
            _msrX = _msrY = 0;
        }
    }

    private void readFRMIsec(IntelDataInputStream inNum)
        throws IOException {
        _frameCnt = inNum.readIntelInt();
        _timeStamp = inNum.readIntelInt();
        _errCode = inNum.readIntelInt();
    }

    private void readTRACsec(IntelDataInputStream inNum)
        throws IOException {
        try {
            _msrX = inNum.readIntelInt();
            _msrY = inNum.readIntelInt();
            /*int _area =*/ inNum.readIntelInt();
            /*int _gap =*/ inNum.readIntelInt();
            /*int _mismatch =*/ inNum.readIntelInt();
        } catch (IOException e) {
            throw new IOException("Invalid Tracking point section data");
        }

        _hasMeasurePoints = true;
    }

    private void readNewTRACsec(IntelDataInputStream inNum)
        throws IOException {
        int validTrackSection;

        try {
            _msrX = inNum.readIntelInt();
            _msrY = inNum.readIntelInt();
            /*int _area =*/ inNum.readIntelInt();
            /*int _gap =*/ inNum.readIntelInt();
            /*int _mismatch =*/ inNum.readIntelInt();
            validTrackSection = inNum.readIntelInt();
        } catch (IOException e) {
            throw new IOException("Invalid Tracking point section data");
        }

        if (validTrackSection != 0) {
            _hasMeasurePoints = true;
        } else {
            _hasMeasurePoints = false;
        }
    }

    private void readDCNTsec(IntelDataInputStream inNum)
        throws IOException {
        /*int _goodPts =*/ inNum.readIntelInt();
        /*int _badPts =*/ inNum.readIntelInt();
        /*int _stretchedPts =*/ inNum.readIntelInt();
        /*int _DMACnt =*/ inNum.readIntelInt();
        /*int _matchCnt =*/ inNum.readIntelInt();
    }

    private void readPROFsec(IntelDataInputStream inNum)
        throws IOException {
        int pointsCnt = inNum.readIntelInt();

        Vector<Float> prof = new Vector<Float>();

        try {
            for (int i = 0; i != pointsCnt; ++i) {
                int x = inNum.readIntelInt();
                int y = inNum.readIntelInt();
                prof.add((float) x);
                prof.add((float) y);
            }
        } catch (IOException e) {
            throw new IOException("Invalid Profile section data");
        }

        _profile = new float[prof.size()];

        for (int i = 0; i != prof.size(); ++i) {
            _profile[i] = prof.elementAt(i);
        }
    }

    private void readIPRFsec(byte[] data, IntelDataInputStream inNum,
        int offset, int lenght) throws IOException {
        try {
            int width = inNum.readIntelInt();
            int height = inNum.readIntelInt();

            Vector<Float> prof = new Vector<Float>();
            byte[] imgData = new byte[width * height];

            ByteArrayInputStream in = new ByteArrayInputStream(data,
                    offset + (2 * 4), lenght - 8);
            readAll(in, imgData, imgData.length);

            for (int i = 0; i < width; i++) {
                for (int j = 0; j < height; j++) {
                    byte point = imgData[(j * width) + i];

                    if (point != 0) {
                        prof.add((float) i);
                        prof.add((float) j);
                    }
                }
            }

            _profile = new float[prof.size()];

            for (int i = 0; i != prof.size(); ++i) {
                _profile[i] = prof.elementAt(i);
            }
        } catch (IOException e) {
            throw new IOException("Invalid IPRF section data");
        }
    }

    private void readFITPsec(IntelDataInputStream inNum)
        throws IOException {
        try {
            _fitPts = inNum.readIntelInt();
            _fit = new float[_fitPts * 2];

            for (int i = 0; i != _fitPts; ++i) {
                _fit[i * 2] = inNum.readIntelInt();
                _fit[(i * 2) + 1] = inNum.readIntelInt();
            }
        } catch (IOException e) {
            throw new IOException("Invalid Fit points section data");
        }
    }

    private void readPLNSsec(IntelDataInputStream inNum)
        throws IOException {
        try {
            _fitPts = inNum.readIntelInt();

            Vector<Float> fit = new Vector<Float>();

            for (int i = 0; i != _fitPts; ++i) {
                int x = inNum.readIntelInt();
                int y = inNum.readIntelInt();

                if (y != 0) {
                    fit.add((float) x);
                    fit.add((float) y);
                }
            }

            _fit = new float[fit.size()];

            for (int i = 0; i != fit.size(); ++i) {
                _fit[i] = fit.elementAt(i);
            }
        } catch (IOException e) {
            throw new IOException("Invalid Fit points section data");
        }
    }

    private double[] readPointsTypeSec(IntelDataInputStream inNum)
        throws IOException {
        double[] secPoints;

        try {
            int pointsCnt = inNum.readIntelInt();
            Vector<Double> points = new Vector<Double>();

            for (int i = 0; i != pointsCnt; ++i) {
                int x = inNum.readIntelInt();
                int y = inNum.readIntelInt();
                points.add((double) x);
                points.add((double) y);
            }

            secPoints = new double[points.size()];

            for (int i = 0; i != points.size(); ++i) {
                secPoints[i] = points.elementAt(i);
            }
        } catch (IOException e) {
            throw new IOException("Invalid 2D section data");
        }

        return secPoints;
    }

    private ImageRepresentation readImgSec(byte[] data,
        IntelDataInputStream inNum, int offset, int lenght)
        throws IOException {
        ImageRepresentation image;
        int width;
        int height;

        try {
            image = new ImageRepresentation();
            image.width = inNum.readIntelInt();
            image.height = inNum.readIntelInt();
            image.bitDepth = inNum.readIntelInt();
            width = image.width;
            height = image.height;
        } catch (IOException e) {
            throw new IOException("Invalid Image section data");
        }

        byte[] img;

        if (image.bitDepth > 8) {
            img = new byte[width * height * 2];
        } else {
            img = new byte[width * height];
        }

        ByteArrayInputStream in = new ByteArrayInputStream(data,
                offset + (3 * 4), lenght - 12);
        readAll(in, img, img.length);
        image.data = img;

        return image;
    }

    private HistogramRepresentation readHistSec(IntelDataInputStream inNum)
        throws IOException {
        HistogramRepresentation hist = new HistogramRepresentation();

        try {
            hist.lenght = inNum.readIntelInt();

            int[] getHist = new int[hist.lenght];

            for (int i = 0; i != getHist.length; ++i) {
                getHist[i] = inNum.readIntelInt();
            }

            hist.histData = getHist;
        } catch (IOException e) {
            throw (new IOException("Invalid Histogram section data"));
        }

        return hist;
    }

    private HistogramRepresentation readHistPointSec(IntelDataInputStream inNum)
        throws IOException {
        try {
            int pointsCnt = inNum.readIntelInt();
            HistogramRepresentation hist = new HistogramRepresentation();
            hist.lenght = pointsCnt * 2;

            int[] data = new int[pointsCnt * 2];

            for (int i = 0; i < pointsCnt; i++) {
                int x = inNum.readIntelInt();
                int high = inNum.readIntelInt();
                data[2 * i] = x;
                data[(2 * i) + 1] = high;
            }

            hist.histData = data;

            return hist;
        } catch (IOException e) {
            throw (new IOException("Invalid Histogram section data"));
        }
    }

    private void readSTTSsec(IntelDataInputStream inNum)
        throws IOException {
        try {
            int size = inNum.readIntelInt();
            _numericData = new double[size];

            for (int i = 0; i < size; i++) {
                _numericData[i] = inNum.readIntelInt();
            }
        } catch (IOException e) {
            throw new IOException("Invalid STTS section data");
        }
    }

    private void readAll(InputStream in, byte[] buf, int len)
        throws IOException {
        int read = 0;
        int r;

        while ((read < len) && ((r = in.read(buf, read, len - read)) != -1)) {
            read += r;
        }
    }

    public float[] getProfile() {
        return _profile;
    }

    public int[] getMeasurePoint() {
        if (_hasMeasurePoints) {
            int[] res = { _msrX, _msrY };

            return res;
        } else {
            return null;
        }
    }

    public float[] getFit() {
        return _fit;
    }

    public double[][] getPoints() {
        return _points;
    }

    public double[] getNumericData() {
        return _numericData;
    }

    public int getTimeStamp() {
        return _timeStamp;
    }

    public int getFrameCnt() {
        return _frameCnt;
    }

    public int getErrorCode() {
        return this._errCode;
    }

    public ImageRepresentation[] getImages() {
        return (ImageRepresentation[]) _images.toArray(new ImageRepresentation[0]);
    }

    public HistogramRepresentation[] getHistograms() {
        return (HistogramRepresentation[]) _histograms.toArray(new HistogramRepresentation[0]);
    }

    public byte[] getFullPackage() {
        return this._fullPackage;
    }
}


class SectionProperties {
    public String pkgSecType;
    public int pkgTabIdx;
    public int idx;
    public String color;
    public int lines;
    public String name;
    public String type;
}
