package com.nts.isip.tcp;

import com.nts.isip.tcp.isipgui.ImageDisplayPanel;

import ij.ImagePlus;

import ij.process.ByteProcessor;
import ij.process.ColorProcessor;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;


/**
 * Thread that dynamically renders images from data packages.
 * The PainterThread is implemented to realize parallel rendering of
 * images data for better performance.
 * The frame image data, retrieved from server, is passed to the setData
 * method, which notifies the waiting thread which starts rendering the data.
 * Once rendered, the image is set to be displayed on the associated ImagePanel,
 * which is passed on construction.
 */
public class PainterThread extends java.lang.Thread {
    private ImageDisplayPanel _imgPanel;
    private byte[] _imgData;
    private Dimension _imgDim = new Dimension();
    private boolean _dataSet = false;
    private int _tabIdx;
    private int _dataIdx;
    private String _channel;
    private double[] _pixelProfile;

    public PainterThread(ImageDisplayPanel panel) {
        _imgPanel = panel;
    }

    public PainterThread(ImageDisplayPanel panel, int tabIdx, int dataIdx,
        String channel) {
        _imgPanel = panel;
        _tabIdx = tabIdx;
        _dataIdx = dataIdx;
        _channel = channel;
    }

    public void run() {
        while (true) {
            synchronized (this) {
                while (!_dataSet) {
                    try {
                        this.wait();
                    } catch (InterruptedException e) {
                    }
                }

                if (_imgData != null) {
                    ByteProcessor bp = new ByteProcessor(_imgDim.width,
                            _imgDim.height);
                    bp.setPixels(_imgData);

                    Image img = new ImagePlus("img", bp).getImage();
                    ColorProcessor cp = new ColorProcessor(img);

                    if (_pixelProfile != null) {
                        for (int i = 0; i < _pixelProfile.length; i += 2) {
                            cp.putPixel((int) _pixelProfile[i],
                                (int) _pixelProfile[i + 1], Color.blue.getBlue());
                        }
                    }

                    _imgPanel.setImage(new ImagePlus("img", cp).getImage());
                } else {
                    _imgPanel.setImage(null);

                    //_imgPanel.setPixelProfile(null);
                }

                _dataSet = false;
            }
        }
    }

    public synchronized void setData(byte[] data, int bitDepth, int width,
        int height, double[] pixelProfile) {
        if (_imgPanel.getDependVisibleParent().isVisible()) {
            _imgDim.width = width;
            _imgDim.height = height;
            _pixelProfile = pixelProfile;

            //_pixelProfile = new double[1000];
            //for(int i = 0; i < 1000; i++) {
            //	_pixelProfile[i] = (double)data[i];
            //}
            if (data == null) {
                _imgData = null;
            } else {
                if (bitDepth <= 8) {
                    _imgData = new byte[data.length];
                    System.arraycopy(data, 0, _imgData, 0, _imgData.length);
                } else {
                    _imgData = new byte[data.length / 2];

                    for (int i = 0; i != _imgData.length; ++i) {
                        int low = data[i * 2] & 0xff;
                        int high = data[(i * 2) + 1] & 0xff;
                        short val = (short) ((high << 8) | low);
                        _imgData[i] = (byte) (val >> (bitDepth - 8));
                    }
                }
            }

            _dataSet = true;
            notify();
        }
    }

    public synchronized int getTabIdx() {
        return this._tabIdx;
    }

    public synchronized int getDataIdx() {
        return this._dataIdx;
    }

    public synchronized String getChannel() {
        return this._channel;
    }

    public synchronized void setTabIdx(int tabIdx) {
        _tabIdx = tabIdx;
    }

    public synchronized void setDataIdx(int dataIdx) {
        _dataIdx = dataIdx;
    }

    public synchronized void setChannel(String channel) {
        _channel = channel;
    }
}
