package com.nts.isip.tcp;

import com.nts.isip.tcp.exceptions.CheckSumException;
import com.nts.isip.tcp.exceptions.FrameSourceException;
import com.nts.isip.tcp.listeners.ProcessorStateListener;
import com.nts.isip.tcp.listeners.StreamSourceStateListener;
import com.nts.isip.tcp.listeners.StreamStateListener;
import com.nts.isip.tcp.processors.FrameProcessorBase;
import com.nts.isip.tcp.sources.FrameSource;
import com.nts.isip.tcp.sources.FrameSourceBase;
import com.nts.isip.tcp.sources.IcamFrameSource;

import java.io.IOException;

import java.util.HashSet;

import javax.swing.JOptionPane;


/**
 * This thread, when activated, continuously retrieves data from the server and displays it in the client view
 * It is invoked and suspended with the startStream and stopStream methods.
 * When active, it sends requests to the server for frame data, and invokes the view object
 * (passed at construction) to display the frame data.
 */
public class StreamThread extends java.lang.Thread
    implements StreamSourceStateListener {
    private boolean _active = false;
    private FrameSource _strSource;
    private FrameSource _auxillarySource;
    private ClientView _view;
    private ClientFrame _parentFrame;
    private FrameProcessorBase _processor;
    private boolean _canGetFrame = false;
    private FrameSource _targetFrameSource = null;
    private int _numFrmsForRec = 0;
    private boolean _frameRecording = false;
    private HashSet<StreamStateListener> _listeners = new HashSet<StreamStateListener>();
    private HashSet<StreamSourceStateListener> _strSourceListeners = new HashSet<StreamSourceStateListener>();
    private HashSet<ProcessorStateListener> _processorListeners = new HashSet<ProcessorStateListener>();

    public StreamThread(FrameSource strSource, ClientFrame parentFrame,
        FrameProcessorBase processor) {
        _strSource = strSource;
        //_view = view;
        _parentFrame = parentFrame;
        _processor = processor;
    }

    public void setView(ClientView view) {
        _view = view;
        _view.getTopFrame()
             .setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
    }

    public ClientView getView() {
        return _view;
    }

    /**
     * Activates the stream. When the stream is active, the thread will asynchroniously
     * request frame data from the server through the connection, and display it on the
     * view
     * @param calc Not used
     */
    public boolean isThreadActive() {
        return _active;
    }

    public synchronized void startStream(boolean calc) {
        _active = true;
        this.notify();
        _parentFrame.setStatus("Starting stream...");
    }

    /**
     * Deactivates the stream. When the stream is not active, it waits in idle state
     * and no invokes no communication.
     *
     */
    public synchronized void stopStream() {
        _active = false;
        this.interrupt();
        _parentFrame.setStatus("Stopping stream...");
    }

    public synchronized void addStateListener(StreamStateListener listener) {
        _listeners.add(listener);
    }

    public synchronized void addProcessorListenere(
        ProcessorStateListener listener) {
        _processorListeners.add(listener);
    }

    public void run() {
        while (true) {
            synchronized (this) {
                while (!_active && !_canGetFrame) {
                    try {
                        this.wait();
                    } catch (InterruptedException e) {
                    }
                }

                notifyListeners();
                _parentFrame.setStatus("Stream started");
            }

            try {
                while (_active || _canGetFrame) {
                    if (_active) {
                        if (_strSource instanceof IcamFrameSource) {
                            _strSource.getLastFrame();
                        } else {
                            _strSource.getNextFrame();
                        }
                    }

                    synchronized (this) {
                        if (!_strSource.getFrameAvailable()) {
                            try {
                                this.wait();
                            } catch (InterruptedException e) {
                            }
                        }
                    }

                    if (_auxillarySource != null) {
                        _strSource.putFrame(_auxillarySource);
                    }

                    synchronized (this) {
                        //_strSource.loadFrame();
                        if (_frameRecording) {
                            if (_targetFrameSource != null) {
                                _targetFrameSource.putFrame(_strSource);
                            }
                        } else if ((_targetFrameSource != null) &&
                                (_numFrmsForRec > 0)) {
                            _targetFrameSource.putFrame(_strSource);
                            _numFrmsForRec--;

                            if (_numFrmsForRec == 0) {
                                _strSource.endRecording();
                            }
                        }
                    }

                    int total_frames = _strSource.getTotalFrames();
                    int current_frame = _strSource.getFrameNumber();
                    String frame_name = _strSource.getFrameName();
                    _view.displayFrameNumbers(total_frames, current_frame);

                    _processor.processFrame((FrameSourceBase) _strSource);

                    FrameSourceBase frameSource = (FrameSourceBase) _strSource;
                    //_processor.addStatusResults(frameSource.getStatus());
                    _processor.addImagesResults(frameSource.getImages());
                    _processor.addHistogramsResults(frameSource.getHistograms());
                    _processor.add2DResults(frameSource.get2DData());
                    _processor.add3DResults(frameSource.get3DData());
                    _processor.addFrameNumberResults(current_frame);
                    //_processor.addTimeStampResults(frameSource.getTimeStamp());
                    //_processor.addNumericDataResults(frameSource.getNumericData());
                    //_processor.addNumericDataNamesResults(frameSource.getNumericDataNames());
                    _processor.addLightLevelResults(frameSource.getLightLevel());
                    _processor.addChannelPropsResults(frameSource.getChannelProps());
                    _processor.addPixelProfileResults(frameSource.getPixelProfile());
                    _processor.addAdminMode(frameSource.isInAdminMode());

                    try {
                        _view.display2DData(_processor.get2DData());
                    } catch (NullPointerException e) {
                    }

                    PropertyRepresentation[] channelProps = _processor.getChannelProps();

                    if (channelProps != null) {
                        _view.displayChannelProps(channelProps);
                    }

                    int[] lld = _processor.getLightLevel();

                    if (lld != null) {
                        _view.displayLightLevel(lld);
                    }

                    //_view.displayStats(_processor.getNumericData());
                    _view.displayFrame(_processor.getFrameNumber(), frame_name);

                    ImageRepresentation[] images = _processor.getImages();

                    try {
                        _view.displayImages(images, _processor.getPixelProfile());
                    } catch (NullPointerException e) {
                    }

                    try {
                        HistogramRepresentation[] histograms = _processor.getHistograms();
                        _view.displayHistograms(histograms);
                    } catch (NullPointerException e) {
                    }

                    try {
                        Thread.sleep(1000);
                    } catch (java.lang.InterruptedException err) {
                    }

                    _canGetFrame = false;
                }
            } /*catch (ProtocolError e) {
              JOptionPane.showMessageDialog(_view.getTopFrame(),
                  "An error has occured while communicating with server.\n" +
                  e.getMessage(), "Communications error",
                  JOptionPane.ERROR_MESSAGE);
            } catch (ServerError e) {
              JOptionPane.showMessageDialog(_view.getTopFrame(),
                  "An error has occured while communicating with server.\n" +
                  e.getMessage(), "Communications error",
                  JOptionPane.ERROR_MESSAGE);
            }*/
            catch (IOException e) {
                JOptionPane.showMessageDialog(_view.getTopFrame(),
                    "A network error has occured.\nThe system told us: " +
                    e.getMessage(), "Network error", JOptionPane.ERROR_MESSAGE);
            } catch (CheckSumException e) {
                JOptionPane.showMessageDialog(_view.getTopFrame(),
                    "A data error has occured.\nThe system told us: " +
                    e.getMessage(), "Data error", JOptionPane.ERROR_MESSAGE);
            } catch (IllegalArgumentException e) {
                JOptionPane.showMessageDialog(_view.getTopFrame(),
                    "A data error has occured.\nThe Processor requires " +
                    " images of type 8-bit, 16-but grayscale or Color images",
                    "Data error", JOptionPane.ERROR_MESSAGE);
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                _active = false;
                notifyListeners();
                _strSource.endRecording();
                _parentFrame.setStatus("Stream stopped");
            }
        }
    }

    private void notifyListeners() {
        java.util.Iterator<StreamStateListener> it = _listeners.iterator();

        while (it.hasNext()) {
            it.next().streamStateChanged(_active);
        }
    }

    private void notifyProcessorListeners() {
        java.util.Iterator<ProcessorStateListener> it = _processorListeners.iterator();

        while (it.hasNext()) {
            it.next().processorSourceChanged();
        }
    }

    public void setStreamSource(FrameSource strSource) {
        if (strSource != null) {
            strSource.removeStreamSourceStateListener(this);
            _strSourceListeners.remove(this);
        }

        strSource.addStreamSourceStateListener(this);
        _strSource = strSource;
        streamSourceChanged(_strSource.getFrameAvailable());
        _strSourceListeners.add(this);
        notifyListeners();
    }

    public FrameSourceBase getStreamSource() {
        return (FrameSourceBase) _strSource;
    }

    public void addStreamSourceStateListener(StreamSourceStateListener listener) {
        if (_strSource != null) {
            _strSource.addStreamSourceStateListener(listener);
        }

        _strSourceListeners.add(listener);
    }

    public synchronized void streamSourceChanged(boolean active) {
        this.notify();
    }

    public void setFrameProcessor(FrameProcessorBase processor) {
        _processor = processor;
        notifyProcessorListeners();
    }

    public FrameProcessorBase getFrameProcessor() {
        return (FrameProcessorBase) _processor;
    }

    public void getFirstFrame()
        throws IOException, IllegalArgumentException, CheckSumException {
        _strSource.getFirstFrame();

        if (_auxillarySource != null) {
            _auxillarySource.getFirstFrame();
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void getPrevFrame()
        throws IOException, IllegalArgumentException, CheckSumException {
        _strSource.getPrevFrame();

        if (_auxillarySource != null) {
            _auxillarySource.getPrevFrame();
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void getNextFrame()
        throws IOException, IllegalArgumentException, CheckSumException {
        _strSource.getNextFrame();

        if (_auxillarySource != null) {
            _auxillarySource.getNextFrame();
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void getLastFrame()
        throws IOException, IllegalArgumentException, CheckSumException {
        _strSource.getLastFrame();

        if (_auxillarySource != null) {
            _auxillarySource.getLastFrame();
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void reloadFrame() throws IOException, CheckSumException {
        _strSource.reloadFrame();

        if (_auxillarySource != null) {
            _auxillarySource.reloadFrame();
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void getFrameByTime(int timestamp)
        throws IOException, CheckSumException {
        _strSource.getFrameByTime(timestamp);

        if (_auxillarySource != null) {
            _auxillarySource.getFrameByTime(timestamp);
        }

        _canGetFrame = true;

        synchronized (this) {
            this.notify();
        }
    }

    public void systemStore() throws IOException {
        if (_strSource != null) {
            _strSource.systemStore();
        }
    }

    public void templateStore(String key)
        throws IOException, FrameSourceException {
        if (_strSource != null) {
            _strSource.templateStore(key);
        }
    }

    public String deleteTemplate(String key) throws IOException {
        if (_strSource != null) {
            String res = _strSource.deleteTemplate(key);

            return res;
        } else {
            throw new IOException("Disconnected");
        }
    }

    public void setTargetFrameSource(FrameSource targetFrameSource) {
        _targetFrameSource = targetFrameSource;
    }

    public FrameSource getTargetFrameSource() {
        return _targetFrameSource;
    }

    public synchronized void setNumberFramesForRecord(int numFrmsForRec) {
        if (_targetFrameSource != null) {
            _numFrmsForRec = numFrmsForRec;
            _strSource.startRecording(_numFrmsForRec);
        }
    }

    public void setFrameRecording(boolean frameRecording) {
        _frameRecording = frameRecording;
    }

    public void setAuxillarySource(FrameSource auxillarySource) {
        _auxillarySource = auxillarySource;
    }

    public FrameSource getAuxillarySource() {
        return _auxillarySource;
    }
}
