package com.nts.isip.tcp.streams;

import java.io.DataInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;


/**
 * Input stream converting little-endian numbers
 *
 */
public class IntelDataInputStream extends DataInputStream {
    public IntelDataInputStream(InputStream in) {
        super(in);
    }

    /**
     * Reads a signed 16-bit number from this data input stream. The
     * method reads two bytes from the underlying input stream. If the two
     * bytes read, in order, are <code>b1</code> and <code>b2</code>,
     * where each of the two values is between <code>0</code> and
     * <code>255</code>, inclusive, then the result is equal to:
     * <ul><code>
     *  (short)((b1 &lt;&lt; 8) | b2)
     * </code></ul>
     * <p>
     * This method blocks until the two bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next two bytes of this input stream, interpreted as a
     *    signed 16-bit number.
     * @exception EOFException if this input stream reaches the end before
     *      reading two bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.FilterInputStream#in
     */
    public final short readIntelShort() throws IOException {
        InputStream in = this.in;
        int ch2 = in.read();
        int ch1 = in.read();

        if ((ch1 | ch2) < 0) {
            throw new EOFException();
        }

        return (short) ((ch1 << 8) + (ch2 << 0));
    }

    /**
     * Reads an unsigned 16-bit number from this data input stream. This
     * method reads two bytes from the underlying input stream. If the
     * bytes read, in order, are <code>b1</code> and <code>b2</code>,
     * where <code>0&nbsp;&lt;=&nbsp;b1</code>,
     * <code>b2&nbsp;&lt;=&nbsp;255</code>, then the result is equal to:
     * <ul><code>
     *  (b1 &lt;&lt; 8) | b2
     * </code></ul>
     * <p>
     * This method blocks until the two bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next two bytes of this input stream, interpreted as an
     *    unsigned 16-bit integer.
     * @exception EOFException if this input stream reaches the end before
     *      reading two bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.FilterInputStream#in
     */
    public final int readIntelUnsignedShort() throws IOException {
        InputStream in = this.in;
        int ch2 = in.read();
        int ch1 = in.read();

        if ((ch1 | ch2) < 0) {
            throw new EOFException();
        }

        return ((ch1 << 8) + (ch2 << 0));
    }

    /**
     * Reads a signed 32-bit integer from this data input stream. This
     * method reads four bytes from the underlying input stream. If the
     * bytes read, in order, are <code>b1</code>, <code>b2</code>,
     * <code>b3</code>, and <code>b4</code>, where
     * 0&nbsp;&lt;=&nbsp;<code>b1</code>, <code>b2</code>,
     * <code>b3</code>, <code>b4</code>&nbsp;&lt;=&nbsp;255, then the
     * result is equal to:
     * <ul><code>
     *  (b1 &lt;&lt; 24) | (b2 &lt;&lt; 16) + (b3 &lt;&lt; 8) +b4
     * </code></ul>
     * <p>
     * This method blocks until the four bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next four bytes of this input stream, interpreted as an
     *    <code>int</code>.
     * @exception EOFException if this input stream reaches the end before
     *      reading four bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.FilterInputStream#in
     */
    public final int readIntelInt() throws IOException {
        //long start = System.currentTimeMillis();
        InputStream in = this.in;
        int ch4 = in.read();
        int ch3 = in.read();
        int ch2 = in.read();
        int ch1 = in.read();

        if ((ch1 | ch2 | ch3 | ch4) < 0) {
            throw new EOFException();
        }

        //long end = System.currentTimeMillis();
        return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));
    }

    /**
     * Reads a signed 64-bit integer from this data input stream. This
     * method reads eight bytes from the underlying input stream. If the
     * bytes read, in order, are <code>b1</code>, <code>b2</code>,
     * <code>b3</code>, <code>b4</code>, <code>b5</code>,
     * <code>b6</code>, <code>b7</code>, and <code>b8</code>, where
     * <ul><code>
     *  0 &lt;= b1, b2, b3, b4, b5, b6, b7, b8 &lt;= 255,
     * </code></ul>
     * <p>
     * then the result is equal to:
     * <p><blockquote><pre>
     *  ((long)b1 &lt;&lt; 56) + ((long)b2 &lt;&lt; 48) +
     *  ((long)b3 &lt;&lt; 40) + ((long)b4 &lt;&lt; 32) +
     *  ((long)b5 &lt;&lt; 24) + (b6 &lt;&lt; 16) +
     *  (b7 &lt;&lt; 8) + b8
     * </pre></blockquote>
     * <p>
     * This method blocks until the eight bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next eight bytes of this input stream, interpreted as a
     *    <code>long</code>.
     * @exception EOFException if this input stream reaches the end before
     *      reading eight bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.FilterInputStream#in
     */
    public final long readIntelLong() throws IOException {
        InputStream in = this.in;
        long ch8 = in.read();
        long ch7 = in.read();
        long ch6 = in.read();
        long ch5 = in.read();
        long ch4 = in.read();
        long ch3 = in.read();
        long ch2 = in.read();
        long ch1 = in.read();

        if ((ch1 | ch2 | ch3 | ch4 | ch5 | ch6 | ch7 | ch8) < 0) {
            throw new EOFException();
        }

        return ((ch1 << 56) | (ch2 << 48) | (ch3 << 40) | (ch4 << 32) |
        (ch5 << 24) | (ch6 << 16) | (ch7 << 8) | (ch8 << 0));
    }

    /**
     * Reads a <code>float</code> from this data input stream. This
     * method reads an <code>int</code> value as if by the
     * <code>readInt</code> method and then converts that
     * <code>int</code> to a <code>float</code> using the
     * <code>intBitsToFloat</code> method in class <code>Float</code>.
     * This method blocks until the four bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next four bytes of this input stream, interpreted as a
     *    <code>float</code>.
     * @exception EOFException if this input stream reaches the end before
     *      reading four bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.DataInputStream#readInt()
     * @see java.lang.Float#intBitsToFloat(int)
     */
    public final float readIntelFloat() throws IOException {
        return Float.intBitsToFloat(readIntelInt());
    }

    /**
     * Reads a <code>double</code> from this data input stream. This
     * method reads a <code>long</code> value as if by the
     * <code>readLong</code> method and then converts that
     * <code>long</code> to a <code>double</code> using the
     * <code>longBitsToDouble</code> method in class <code>Double</code>.
     * <p>
     * This method blocks until the eight bytes are read, the end of the
     * stream is detected, or an exception is thrown.
     *
     * @return the next eight bytes of this input stream, interpreted as a
     *    <code>double</code>.
     * @exception EOFException if this input stream reaches the end before
     *      reading eight bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.DataInputStream#readLong()
     * @see java.lang.Double#longBitsToDouble(long)
     */
    public final double readIntelDouble() throws IOException {
        return Double.longBitsToDouble(readIntelLong());
    }

    /**
     * Reads a <code>String</code> from this data input stream.
     *
     * @param padLength is the amount of characters this string ocuppies
     * @return String that is zero terminated in a padded buffer
     * @exception EOFException if this input stream reaches the end before
     *      reading eight bytes.
     * @exception  IOException   if an I/O error occurs.
     * @see java.io.DataInputStream#readLong()
     * @see java.lang.Double#longBitsToDouble(long)
     */
    public final String readIntelString(int padLength)
        throws IOException {
        InputStream in = this.in;
        StringBuffer temp = new StringBuffer(128);
        char b;

        for (; (padLength != 0) && ((b = (char) in.read()) != 0);
                --padLength)
            temp.append(b);

        for (; --padLength > 0;)
            b = (char) in.read();

        return (temp.toString());
    }
}
